import numpy as np


def check_image(image):
    """Check if the input is a numpy array

    Args:
        image (np.ndarray): Input image to be transformed

    Raises:
        TypeError: If the input is not a numpy array
        ValueError: If the input is not a grayscale or color image
    """
    if not isinstance(image, np.ndarray):
        raise TypeError(
            f"Input should be a numpy array. Expected <class 'numpy.ndarray'>, but got {type(image)}"
        )

    if len(image.shape) not in [2, 3]:
        raise ValueError(
            f"Input should be grayscale or color image. Got {len(image.shape)} dimensions"
        )



import numpy as np
from imgcv.common import check_image


def logarithmic_transform(image, c=1):
    """Apply logarithmic transformation to the input image

    Formula:
        s = c * log(1 + r)

    Args:
        image (ndarray): Input image to be transformed
        c (float): Constant value to scale the logarithmic transformation

    Raises:
        TypeError: If the input is not a numpy array
        ValueError: If the input is not a grayscale or color image

    Returns:
        np.ndarray: Transformed image
    """

    check_image(image)

    # Normalizing the image
    image = image / 255.0

    # Applying the logarithmic transformation
    transformed_img = c * np.log(1 + image)
    transformed_img = (transformed_img * 255.0).astype(np.uint8)

    return transformed_img



from imgcv.common import check_image
import numpy as np
from warnings import warn


def histogram_equalization(img):
    """Equalizes the histogram of the image and returns the equalized image.

    Note: This function currently only works for grayscale images.

    Args:
        img (np.ndarray): Image to equalize the histogram

    Returns:
        np.ndarray: Equalized image
    """

    check_image(img)

    pixel_freqs = calculate_histogram(img)
    pdf = calculate_pdf(img, pixel_freqs)
    cdf = calculate_cdf(img, pdf)

    equi_hist = {}

    for key in cdf:
        equi_hist[key] = round(cdf[key] * 255)

    equ_im = np.zeros(img.shape)

    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
            equ_im[i][j] = equi_hist[img[i][j]]

    return equ_im, equi_hist


def calculate_histogram(img):
    """Calculates the histogram of the image

    Note: This function currently only works for grayscale images.

    Args:
        img (np.ndarrray): Image to calculate the histogram

    Returns:
        dict: Dictionary containing the frequency of each pixel value
    """
    check_image(img)

    # creating dictionary to store the frequency of each pixel value
    pixel_freq = {}

    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
            if img[i][j] in pixel_freq:
                pixel_freq[img[i][j]] += 1
            else:
                pixel_freq[img[i][j]] = 1

    pixel_freq = dict(sorted(pixel_freq.items()))

    return pixel_freq


def calculate_pdf(img, pixel_freqs):
    """Calculates the probability density function of the image

    Note: This function currently only works for grayscale images.

    Args:
        img (np.ndarray): Image to calculate the pdf
        pixel_freqs (dict): Dictionary containing the frequency of each pixel value

    Returns:
        dict: Dictionary containing the probability density function of the image
    """
    check_image(img)

    check_image(img)

    total_pixels = img.shape[0] * img.shape[1]

    # probability density function
    pdf = {}

    for key in pixel_freqs:
        pdf[key] = np.round(pixel_freqs[key] / total_pixels, 4)

    return pdf


def calculate_cdf(img, pdf):
    """Calculates the cumulative density function of the image

    Note: This function currently only works for grayscale images.

    Args:
        img (np.ndarray): Image to calculate the cdf
        pdf (dict): Dictionary containing the probability density function of the image

    Returns:
        dict: Dictionary containing the cumulative density function of the image
    """

    check_image(img)

    # cumulative density function
    cdf = {}

    # each cdf is the sum of current pdf + previous cdf
    for i, key in enumerate(pdf):
        # if it is the first value
        if i == 0:
            cdf[key] = pdf[key]
        else:
            # this is taking the previous cdf value and adding the current pdf value
            cdf[key] = cdf[list(cdf.keys())[i - 1]] + pdf[key]

    return cdf



import numpy as np
from imgcv.common import check_image


def negative_transform(image):
    """Invert the pixel values of the input image

    Args:
        image (ndarray): Input image to be transformed

    Formula:
        s = 255 - r

    Raises:
        TypeError: If the input is not a numpy array
        ValueError: If the input is not a grayscale or color image

    Returns:
        ndarray: Transformed image
    """

    check_image(image)

    # check the image shape
    if len(image.shape) == 2:
        return 255 - image
    elif len(image.shape) == 3:
        # need to tackle each channel separately
        transformed_img = np.zeros_like(image)
        for i in range(image.shape[2]):
            transformed_img[:, :, i] = 255 - image[:, :, i]

        return transformed_img
